# -*- coding: utf-8 -*-
#
# Copyright (c) 2007, 2008, 2011 by Intevation GmbH
#
# FormEd is free software under the terms of the
# GNU General Public License Version 3. See file
# LICENSE coming with the source of FormEd for details.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
# Björn Ricks <bjoern.ricks@intevation.de>

import re
import sys

import formed.meta.data as data
from formed.meta.io.document import saveDocument
from formed.meta.nodecomponents import VISIT_CONTINUE


ORDER_RE = re.compile("([^;]*);(.*)")

DATE_RULE_VALUE = \
u"'%(a_desc)s' darf zeitlich nicht nach '%(b_desc)s' liegen."

DATE_RULE_DESC = \
u"'%(a)s' muss vor '%(b)s' liegen."

DATE_RULE_EXPR = \
u"%(a)s isset %(a)s known and %(b)s isset %(b)s known and and %(a)s %(b)s > and not"

IN_PAST_EXPR = \
u"%(a)s isset %(a)s known and %(a)s today > and not"

IN_PAST_VALUE = \
u"'%(a_desc)s' darf nicht in der Zukunft liegen."

IN_PAST_DESC = \
"'%(a)s' darf nicht in der Zukunft liegen."

IN_FUTURE_EXPR = \
u"%(a)s isset %(a)s known and today %(a)s >= and not"

IN_FUTURE_DESC = \
u"'%(a)s muss in der Zukunft liegen.'"

IN_FUTURE_VALUE = \
u"'%(a_desc)s' darf nicht in der Vergangenheit liegen."


class WidgetCollectorWithRepeatGroups(data.WidgetCollector):

    def handleRepeatNode(self, repeat):
        return VISIT_CONTINUE


class VariableType(object):

    def __init__(self, formed):
        self.formed = formed
        self.entities = {}

    def substitute(self, name, prefix='$', cache=True):
        try:
            return self.entities[name]
        except KeyError:
            nc = self.formed.findByName(name)
            if not nc:
                print >> sys.stderr, "WARNING: node '%s' not found."
                value = "%s%s" % (prefix, name)
            else:
                is_repeat = False
                while nc:
                    if isinstance(nc, data.RepeatNode):
                        is_repeat = True
                        break
                    nc = nc.parent
                value = "%s%s" % (prefix, name)
            if cache:
                self.entities[name] = value
            return value


def traverse_recursive(current, graph, chain, found):
    xchain = chain[:]
    xchain.append(current)

    succs = graph[current][1]

    if not succs:
        found(xchain)
        return

    for succ in succs:
        if succ in chain:
            found(xchain)
        else:
            traverse_recursive(succ, graph, xchain[:], found)


def all_paths(graph):
    # find all starting node (no preds)
    starts = []
    for name, preds_succs in graph.iteritems():
        if not preds_succs[0]:
            starts.append(name)

    paths = []

    def found(chain):
        paths.append(chain)

    for start in starts:
        traverse_recursive(start, graph, [], found)

    return paths


def all_pairs(path):
    N = len(path)
    for i in xrange(N):
        for j in xrange(i + 1, N):
            yield (path[i], path[j])


def descriptions(node):
    out = []
    desc = node.getDescription().strip()
    if desc:
        out.insert(0, desc)
    return out


class GenerateRequiredRules:

    def __init__(self, document):
        self.document = document

    def is_in_repeat_group(self, node, reverse=True):
        """Return a list with all the (nested) repeat groups the node is in
        (walking up to the node)"""
        rg = set()
        for parent in node.pathToRoot():
            if isinstance(parent, data.RepeatNode):
                #rg.add(parent)
                rg.add(parent.getName())
        return rg

    @staticmethod
    def getDescription():
        return u"Generate rules for fields that are required"

    def generateDateSequenceRules(self):
        document = self.document

        var_type = VariableType(document)

        all_dates = list(document.findAllByClass(data.DateLeaf))

        # Iterate through all the dates to generate the requiered rules, taking
        # into cosideration whether they are in a repeat group or not

        rg_graph = {}
        for date in all_dates:
            rg = self.is_in_repeat_group(date, False)
            if rg != set():
                rg_graph[date.getName()] = rg

        date_graph = dict([(d.getName(), (set(), set())) for d in all_dates])


        today_in_future = set()
        today_in_past = set()

        for date in all_dates:
            order = date.getOrder()
            m = ORDER_RE.match(order)
            if not m:
                continue

            name = date.getName()

            preds = m.group(1)
            succs = m.group(2)
            p_set = date_graph[name][0]
            s_set = date_graph[name][1]

            for pred in preds.split(","):
                pred = pred.strip()
                if not pred:
                    continue
                if pred == "$TODAY":
                    today_in_past.add(name)
                    continue
                try:
                    # add current to succs of pred
                    date_graph[pred][1].add(name)
                except KeyError:
                    print >> sys.stderr, "pred date '%s' not found" % pred
                    continue
                # add to preds
                p_set.add(pred)

            for succ in succs.split(","):
                succ = succ.strip()
                if succ == "$TODAY":
                    today_in_future.add(name)
                    continue
                if not succ:
                    continue
                try:
                    # add current to preds of succ
                    date_graph[succ][0].add(name)
                except KeyError:
                    print >> sys.stderr, "succ date '%s' not found" % succ
                    continue
                # add to succs
                s_set.add(succ)

        # consolidate date graph
        for name, preds_succs in date_graph.iteritems():
            for pred in preds_succs[0]:
                date_graph[pred][1].add(name)
            for succ in preds_succs[1]:
                date_graph[succ][0].add(name)

        # eliminate isolated dates
        to_be_deleted = []
        for name, preds_succs in date_graph.iteritems():
            if not preds_succs[0] and not preds_succs[1]:
                print >> sys.stderr, \
                    "Date '%s' is not neighbored to other dates." % name
                to_be_deleted.append(name)

        for name in to_be_deleted:
            del date_graph[name]

        # generate all paths from start to end dates.
        paths = all_paths(date_graph)

        # build all pairs of dates to be checked
        pairs = set()
        for path in paths:
            for pair in all_pairs(path):
                pairs.add(pair)

        dates = dict([(d.getName(), d) for d in all_dates])

        for idx, pair in enumerate(pairs):
            # Generate the rules according to who is in which repeat group.

            # NOTICE: at the moment we are assuming that there are no nested
            # repeat groups
            #
            # For this reason the repeat groups' names are saved in rg_graph, so
            # once nested repeat groups are added to libformed/formedtree this
            # function can be adapted

            # Both in (the same) RG or both outside of a RG (most usual case)
            # both: $
            prefix_rule_1 = '$'
            prefix_rule_2 = '$'

            # 1st Date in a RG, 2nd not
            # 1st: % 2nd: $
            if pair[0] in rg_graph and pair[1] not in rg_graph:
                prefix_rule_1 = '%'
                prefix_rule_2 = '$'

            # 1st Date not in a RG, 2nd in one
            # 1st: $ 2nd: %
            elif pair[0] not in rg_graph and pair[1] in rg_graph:
                prefix_rule_1 = '$'
                prefix_rule_2 = '%'

            rule1 = data.RuleLeaf()
            rule2 = data.RuleLeaf()
            rule1.setName("date-sequence-rule-%d-1" % idx)
            rule2.setName("date-sequence-rule-%d-2" % idx)
            d1, d2 = dates[pair[0]], dates[pair[1]]
            rule1.setMark("error:%s" % d1.getName())
            rule2.setMark("error:%s" % d2.getName())

            value = DATE_RULE_VALUE % {'a_desc': d1.getDescription().strip(),
                                       'b_desc': d2.getDescription().strip()}

            rule1.setValue(value)
            rule2.setValue(value)
            param = {'a': pair[0], 'b': pair[1]}
            rule1.setDescription(DATE_RULE_DESC % param)
            rule2.setDescription(DATE_RULE_DESC % param)
            param = {'a': var_type.substitute(pair[0], prefix_rule_1, False),
                     'b': var_type.substitute(pair[1], prefix_rule_2, False)}

            rule1.setExpr(DATE_RULE_EXPR % param)
            rule2.setExpr(DATE_RULE_EXPR % param)

            d1.getParent().addChild(rule1)
            d2.getParent().addChild(rule2)

        for d in today_in_past.copy():
            stack = [d]
            while stack:
                d = stack.pop()
                today_in_past.add(d)
                try:
                    stack.extend(date_graph[d][1])
                except KeyError:
                    pass

        for idx, d in enumerate(today_in_past):
            rule = data.RuleLeaf()
            rule.setName("date-in-future-rule-%d" % idx)
            rule.setExpr(IN_FUTURE_EXPR % {'a': var_type.substitute(d)})
            rule.setDescription(IN_FUTURE_DESC % {'a': d})
            rule.setMark("error:%s" % d)
            d1 = dates[d]
            rule.setValue(IN_FUTURE_VALUE
                          % {'a_desc': u" / ".join(descriptions(d1))})
            d1.getParent().addChild(rule)

        for d in today_in_future.copy():
            stack = [d]
            while stack:
                d = stack.pop()
                today_in_future.add(d)
                try:
                    stack.extend(date_graph[d][0])
                except KeyError:
                    pass

        for idx, d in enumerate(today_in_future):
            rule = data.RuleLeaf()
            rule.setName("date-in-past-rule-%d" % idx)
            rule.setExpr(IN_PAST_EXPR % {'a': var_type.substitute(d)})
            rule.setDescription(IN_PAST_DESC % {'a': d})
            rule.setMark("error:%s" % d)
            d1 = dates[d]
            rule.setValue(IN_PAST_VALUE
                          % {'a_desc': u" / ".join(descriptions(d1))})
            d1.getParent().addChild(rule)

    def generateRequiredRules(self):
        document = self.document

        var_type = VariableType(document)

        collector = WidgetCollectorWithRepeatGroups()
        document.visit(collector.visitor)
        widgets = collector.widgets
        # Get all rules in document

        # Make a list of list with manually created required rules that have
        # some special logic in them
        known_fields = set()
        for r in document.findAllByClass(data.RuleLeaf):
            if 'required-rule-manual' in r.getName():
                known_fields = known_fields.union(r.getExpr().getDependencies())

        rule_counter = len(known_fields)

        for w in widgets:
            # ignore if the rule is in the manual list
            if w.getName() in known_fields:
                continue

            flags = w.getFlags()
            parent = w.getParent()
            x = var_type.substitute(w.getName())

            if flags.find('required') >= 0:
                if isinstance(w, (data.ChoiceNode, data.RadioNode)):
                    expr = "%s -1 == not" % x
                else:
                    expr = "%s isset" % x

                # create new rule
                rule = data.RuleLeaf()
                rule.setName("required-rule-%d" % rule_counter)
                rule_counter += 1
                rule.setExpr(expr)
                rule.setValue("Fehlende Pflichtfeldangabe")
                rule.setMark("warning:%s" % w.getName())

                # add rules to the tree
                parent.addChild(rule)

            # create additional error rule for datefields
            # Daterules must not be "unknown"
            if flags.find('time') >= 0:
                if isinstance(w, (data.DateLeaf)):
                    expr = "%s known" % x
                    erule = data.RuleLeaf()
                    erule.setName("unknown-phasedate-rule-%d" % rule_counter)
                    erule.setExpr(expr)
                    erule.setDescription("Das Datum darf nicht 'unbekannt' sein.")
                    erule.setValue("Das Datum darf nicht 'unbekannt' sein.")
                    erule.setMark("error:%s" % w.getName())
                    parent.addChild(erule)

    def doExport(self, path):
        self.generateRequiredRules()
        self.generateDateSequenceRules()
        saveDocument(self.document, path)


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf-8 enc=utf-8 :

