# -*- coding: utf-8 -*-
# Copyright 2010 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Bernhard Herzog <bernhard.herzog@intevation.de>

"""Structure instance trees that match the meta structure trees."""

import uuid

from formed.meta.data import RepeatNode
from formed.meta.exprtree import VariableProvider


class StructureInstanceNode(object):

    """Node of the structure instance tree.

    Public read-only instance variables:

       uuid -- The uuid of the node

       id -- The database id of the node.  None if the node is not in
             the database

       meta_node -- The corresponding meta structure node.
    """

    def __init__(self, meta_node, id, uuid=None, leaves=None,
                 repeat_groups=None):
        """Initialize the instance.
        The parameters meta_node, id and uuid are simply bound to the
        corresponding instance variables.  The parameters leaves and
        repeat_groups, if give and not None, should be dictionaries.
        The leaves dictionary maps the names of the leaves of this
        instance to the values of the leaves.  The repeat_groups
        dictionary maps the names of the repeat groups to dictionaries
        mapping the instance ids to the actual repeat group instances.
        The names of the leaves and repeat groups must be the ones
        defined in the meta_node.  All leaves must occur in the leaves
        dictionary.

        If the leaves dictionary is not given or None, a dictionary with
        default values is created.  The default values are taken from
        the formed description.
        """
        self.meta_node = meta_node
        self.id = id
        self.uuid = uuid
        if leaves is None:
            leaves = self.default_leaves()
        self.leaves = leaves
        if repeat_groups is None:
            repeat_groups = self.default_repeat_groups()
        self.repeat_groups = repeat_groups

    def default_leaves(self):
        leaves = dict()
        for leaf in self.meta_node.leaves.values():
            leaves[leaf.getName()] = leaf.findDefault()
        return leaves

    def default_repeat_groups(self):
        repeat_groups = dict()
        for rg_meta_node in self.meta_node.repeat_groups.itervalues():
            instances = repeat_groups[rg_meta_node.name] = dict()
            for i in range(rg_meta_node.meta_node.getMin()):
                node = self.__class__(rg_meta_node, None)
                if node.id is None:
                    node.id = id(node)
                instances[node.id] = node
        return repeat_groups

    def get(self, name):
        local_name = name[0]
        restname = name[1:]

        if isinstance(local_name, tuple):
            item = self.get_repeat_group(*local_name)
            if restname:
                item = item.get(restname)
        else:
            if local_name in self.meta_node.leaves:
                item = self.get_leaf(local_name)
            elif local_name in self.meta_node.repeat_groups:
                item = self.get_repeat_groups(local_name)
            else:
                raise ValueError("Name %r is not known as leaf or repeat group"
                                 % local_name)
        return item

    def set_value(self, name, value):
        local_name = name[0]
        restname = name[1:]

        if isinstance(local_name, tuple):
            item = self.get_repeat_group(*local_name)
            if restname:
                item.set_value(restname, value)
        else:
            meta_leaf = self.meta_node.leaves.get(local_name)
            if meta_leaf is None:
                raise ValueError("Name %r is not known as a leaf" % local_name)
            self.set_leaf(local_name, value)

    def get_leaf(self, meta_name):
        return self.leaves[meta_name]

    def set_leaf(self, meta_name, value):
        self.leaves[meta_name] = value

    def get_repeat_groups(self, rg_name):
        if rg_name not in self.meta_node.repeat_groups:
            raise ValueError("Node %r has no repeat group child %r"
                             % (self.meta_node.name, rg_name))
        return self.repeat_groups.setdefault(rg_name, dict())

    def get_repeat_group(self, rg_name, rg_id):
        return self.get_repeat_groups(rg_name)[rg_id]

    def create_repeat_group(self, rg_name):
        node = self.__class__(self.meta_node.repeat_groups[rg_name], None,
                              uuid=uuid.uuid4())
        node.id = id(node)
        self.get_repeat_groups(rg_name)[node.id] = node
        return node

    def delete_repeat_group(self, rg_name, rg_id):
        del self.get_repeat_groups(rg_name)[rg_id]

    def make_unique(self):
        """Assign new UUIDs to the node and contained repeat groups recursively
        """
        self.uuid = uuid.uuid4()
        for rgs in self.repeat_groups.itervalues():
            for rg in rgs.itervalues():
                rg.make_unique()


def pprint_structure_instance_node(node, indent=""):
    """Recursively pretty-print a structure instance node to stdout"""
    print ("%s%s:%s:%s: %d leaves, %d repeat groups"
           % (indent, node.meta_node.name, node.id, node.uuid,
              len(node.leaves), len(node.repeat_groups)))
    indent = "    " + indent
    for key, item in sorted(node.leaves.items()):
        print "%s%s: %r" % (indent, key, item)
    for name, rgs in sorted(node.repeat_groups.items()):
        for key, item in sorted(rgs.items()):
            pprint_structure_instance_node(item, indent)


class StructureInstanceTree(object):

    """Structure instance tree.

    Public read-only instance variables:

      root -- The root structure instance node

      meta_structure -- The corresponding structure meta tree

      uuid -- The uuid of the tree.  Identical to the uuid of the root
              node.

      id -- The database id of the instance.  Identical to the id of the
            root node.
    """

    def __init__(self, meta_structure, root=None, id=None, uuid=None):
        self.meta_structure = meta_structure
        if root is not None:
            self.root = root
        else:
            self.root = self.instantiate_root(meta_structure.root, id, uuid)

    @property
    def uuid(self):
        return self.root.uuid

    @property
    def id(self):
        return self.root.id

    def instantiate_root(self, meta_root, id, uuid):
        return StructureInstanceNode(meta_root, id, uuid=uuid)

    def get(self, name):
        """Return the value of the item indicated by name.
        The argument name can be of two forms: it may be a string using
        the legacy syntax used by formed.instance.tree, or a new item
        specification.

        If the name is a string containing at least one colon (':'), it
        is assumed to be a legacy name.  Legacy names are converted to
        item specifications and then treated like those.

        If the name is not a legacy name, it's an item specification.
        Item specifications have form (rgid1, rgid2, ..., itemname).
        The last part of that tuple is the name of the item as used in
        the meta tree.  The rgidN values are integers specifying the ids
        of the repeat groups that contain the instance of the item.  The
        meta structure defines how many repeat groups there are in the
        hierarchy above the item, so the number of repeat group ids
        required is always known.

        If itemname ist the name of a repeat group, there are two cases
        to consider, though.  The item specification can refer to all
        instances of the repeat group or to a particular instance.  If
        the the number of repeat group ids given in the item
        specification is equal to the number of the repeat groups above
        the item named by itemname, the specification refers to all
        instances.  If there's one more repeat group id, it is
        interpreted as the id of a particular instance of the repeat
        group named by itemname.

        Another special case is items that are not contained in repeat
        groups.  In this case the item name can be given directly as a
        string.  It does not have to be wrapped in a single-element
        tuple.
        """
        return self.root.get(self.full_name(name))

    def set_value(self, name, value):
        self.root.set_value(self.full_name(name), value)

    def full_name(self, name):
        # name is either simply a string with a field name, or a tuple
        # (rgid1, ..., rgidN, fieldname)
        if isinstance(name, basestring):
            if ":" in name:
                name = self.parse_legacy_name(name)
            else:
                name = (name,)

        path = self.meta_structure.structure_path(name[-1])
        num_ids = len(name) - 1
        path_len = len(path)
        if isinstance(path[-1], RepeatNode):
            if num_ids not in (path_len, path_len - 1):
                raise ValueError("The full name for repeat group %r must have"
                                 " either %d or %d ids."
                                 % (name, path_len - 1, path_len))
        else:
            if num_ids != path_len - 1:
                raise ValueError("The full name for item %r must have exactly "
                                 " %d ids." % (name, path_len - 1))
        full_name = []
        for i in range(num_ids):
            full_name.append((path[i].getName(), name[i]))
        if num_ids < path_len:
            full_name.append(name[-1])
        return full_name

    def parse_legacy_name(self, legacy_name):
        """Convert a legacy instance item name into the new item specification.
        The legacy name is a string consisting of colon-separated
        fields.  The first field is the item's meta name.  The rest of
        the fields are the ids of the repeat groups containing the item,
        starting with the one immediately containing the instance item,
        followed by the id of the repeat group containing that repeat
        group, and so on.  The master table itself is treated as a
        repeat group in this context, so the last repeat group id is
        always the case id.
        """
        assert ":" in legacy_name
        parts = legacy_name.split(":")
        parts.reverse()
        for i in range(len(parts) - 1):
            parts[i] = int(parts[i])
        if parts[0] != self.root.id:
            raise ValueError("Legacy name must contain the case id")
        return tuple(parts[1:])

    def get_all(self, name):
        """Return all values for the leaf named by name.
        The return value is a sequence of tuples of the form (NAME,
        VALUE) where NAME is a name usable for the get method to
        retrieve the value and VALUE is the current value of the item.
        """
        def generate_items(node, path):
            child_meta = path[0]
            child_name = child_meta.getName()
            if isinstance(child_meta, RepeatNode):
                for rg in node.get_repeat_groups(child_name).values():
                    for partial_id, value in generate_items(rg, path[1:]):
                        yield ((rg.id,) + partial_id, value)
            else:
                yield ((child_name,), node.get((child_name,)))

        path = self.meta_structure.structure_path(name)

        return list(generate_items(self.root, path))

    def iter_all_items(self):
        """Yield all values in the tree as (key, value) pairs."""
        nodes = [self.meta_structure.root]
        while nodes:
            node = nodes.pop(0)
            for name in node.leaves:
                for item in self.get_all(name):
                    yield item
            nodes.extend(node.repeat_groups.values())

    def make_unique(self):
        """Assign new UUIDs to all nodes recursively."""
        self.root.make_unique()


class StructureVariableProvider(VariableProvider):

    def __init__(self, tree, reference, shadow=None):
        self.tree = tree

        if reference is None:
            self.reference_ids = ()
            self.reference_parent = None
        else:
            self.reference_ids = reference[:-1]
            self.reference_parent = self._structure_parent(reference[-1])

        if shadow is None:
            shadow = dict()
        self.shadow = shadow

    def _structure_parent(self, name):
        for nc in reversed(self.tree.meta_structure.structure_path(name)):
            if isinstance(nc, RepeatNode):
                return nc
        return None

    def getValueForVar(self, name):
        key = (name,)
        if self.reference_parent is self._structure_parent(name):
            key = self.reference_ids + key
        if key in self.shadow:
            return (key, self.shadow[key])
        return (key, self.tree.get(key))

    def getAllValuesForVar(self, name):
        return [(key, self.shadow.get(key, value))
                for key, value in self.tree.get_all(name)]
