#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2011 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>

"""%prog [options] expr1 [expr2 ...]

Evaluate one or more formed expressions. Unless the --dump-tree option
is given, each of the expressions is evaluated and the repr of the
resulting Python object is printed to stdout.
"""

import re
from optparse import OptionParser

from formed.meta.exprtree import Expr, VariableProvider as BaseVariableProvider


class VariableProvider(BaseVariableProvider):

    def __init__(self, local_variables):
        self.local_variables = local_variables

    def getValueForVar(self, var):
        return (var, self.local_variables[var])


def evaluate(expression, boolean, variables):
    expr = Expr(expression)
    result = expr.evaluate(VariableProvider(variables))
    if boolean:
        result = bool(result)
    return result


def expression_tree(expression, boolean, variables):
    expr = Expr(expression)
    expr.compile()
    return expr.prog


def parse_variables(raw_variables):
    variables = dict()
    for raw in raw_variables:
        match = re.match("([a-zA-Z_]+)=", raw)
        if match:
            expr = raw[match.end(0):].strip()
            if expr:
                value = evaluate(expr, False, variables)
            else:
                value = None
            variables[match.group(1)] = value
        else:
            raise ValueError("Invalid variable definition %r", raw)
    return variables


def main():
    parser = OptionParser(usage=__doc__)
    parser.set_defaults(var=[])
    parser.add_option("--bool", action="store_true",
                      help=("Interpret the result of the expression as a"
                            " boolean value."))
    parser.add_option("--dump-tree", action="store_true",
                      help=("Print a dump of the tree representation of the"
                            " expressions instead of evaluating it."))
    parser.add_option("--var", action="append",
                      help=("Define a variable that the expressions can refer"
                            " to. The parameter should have"
                            " the form VARNAME=EXPR. The EXPR part will"
                            " be evaluated and the result assigned to the"
                            " variable given by VARNAME. VARNAME must"
                            " contain only letters and underscores."
                            " The EXPR may refer to variables defined"
                            " previously. If EXPR is omitted, the value will"
                            " be None so that the isset operator will evaluate"
                            " to false for this variable."))

    options, rest = parser.parse_args()

    variables = parse_variables(options.var)

    function = evaluate
    if options.dump_tree:
        function = expression_tree

    for expression in rest:
        value = function(expression, options.bool, variables)
        print repr(value)

main()
