"""Usage: %prog [options] formedtree.xml
Print the structure tree of formedtree.xml in the canonical text form.
"""

import sys
import codecs
from optparse import OptionParser

from formed.meta.io.document import openDocument, namesToClasses
from formed.meta.structure import StructureTree
from formed.meta.data import ChoiceNode


# Map formed meta class names to formed element names. It's a reverse of
# the namesToClasses mapping used when reading the forme meta tree.
class_to_name = dict((cls, name) for name, cls in namesToClasses.iteritems())


def sort_name_key(obj):
    """Return the name of obj.
    Intended to be used as the key argument of the sorted builtin"""
    get_name = getattr(obj, "getName", None)
    if get_name is not None:
        return get_name()
    return obj.name


def sort_description_key(obj):
    """Return the a tuple (description, name) for obj.
    Intended to be used as the key argument of the sorted builtin.

    The name is included in the sort key, so that fields with very
    common descriptions (such as "Sonstiges") are still sorted in a
    consistent way.
    """
    return (obj.getDescription(), obj.getName())



def print_canonical_structure_node(out, node, sort_key, indent=""):
    """Print the structure node node to out."""
    for leaf in sorted(node.leaves.values(), key=sort_key):
        out.write("%s%s: %s: %s\n"
                  % (indent, leaf.getName(), class_to_name[leaf.__class__],
                     leaf.getDescription()))
        out.write("%s  anonymization: %s\n"
                  % (indent, leaf.getAttribute("function")))
        if isinstance(leaf, ChoiceNode):
            for key, desc in leaf.get_options():
                out.write("%s%5d: %s\n" % (indent, int(key), desc))
    for name, rg in sorted(node.repeat_groups.items()):
        out.write("%s%s: repeatgroup:\n" % (indent, name))
        print_canonical_structure_node(out, rg, sort_key,
                                       indent=indent + "    ")


def print_canonical_spec(structure_tree, sort_key):
    """Print the canonical text representation of the structure node"""
    out = codecs.getwriter("utf-8")(sys.stdout)
    print_canonical_structure_node(out, structure_tree.root, sort_key)


def main():
    parser = OptionParser(usage=__doc__)
    parser.set_defaults(sort_by="name")
    parser.add_option("--sort-by",
                      help=("The attribute to sort the fields by"
                            " (e.g. name, description). Default: name"))

    options, rest = parser.parse_args()

    if not rest:
        print >>sys.stderr, "Missing formed filename"
        sys.exit(1)
    formed_file = rest[0]

    if options.sort_by == "name":
        sort_key = sort_name_key
    elif options.sort_by == "description":
        sort_key = sort_description_key
    else:
        print >>sys.stderr, "Unknown --sort-by option %r" % (options.sort_by,)
        sys.exit(1)

    formed = openDocument(formed_file)
    structure_tree = StructureTree(formed)
    print_canonical_spec(structure_tree, sort_key)

main()
