from datetime import date

import xml.etree.ElementTree as ET

LIBFORMED_XML_CONTAINER_VERSION = "5"

XML_ENCODERS = {
    unicode: lambda s: s,
    date: lambda s: "%04d-%02d-%02d" % (s.year, s.month, s.day)
}


def serialize_case(instance, external=None):
    """Return the structure instance tree as an XML string.
    If the parameter external is given and not None, it will be treated
    as a callable which is called with the ElementTree.Element instance
    containing the XML representation of the case.  The callable may
    modify the element instance to add more information, like the
    logbook entries associated with the case.
    """
    formed = instance.meta_structure.meta_tree
    cases = create_cases_container(formed.getVersion())
    case = add_case(cases, instance)
    if external is not None:
        external(case)
    xml = ET.ElementTree(cases)
    return ET.tostring(xml.getroot())


def create_cases_container(formed_type):
    return ET.Element("cases",
                      {"version": LIBFORMED_XML_CONTAINER_VERSION,
                       "formed-type": formed_type})


def add_case(cases, instance):
    """Adds the serialized simplified instance to the cases element"""
    case = ET.SubElement(cases, "case")
    add_node(case, instance.root)
    return case


def add_node(parent, node):
    meta_node = node.meta_node
    element_name = meta_node.name
    if element_name is None:
        element_name = "master"
    element = ET.SubElement(parent, element_name)
    add_leaf(element, "uuid_id", node.uuid)
    for nc in meta_node.ordered_leaves:
        name = nc.getName()
        add_leaf(element, name, node.get((name,)))
    for meta_rg in meta_node.ordered_repeat_groups:
        for rg_instance in node.get_repeat_groups(meta_rg.name).itervalues():
            add_node(element, rg_instance)

def add_leaf(element, name, value):
    if value is not None:
        leaf_element = ET.SubElement(element, name)
        leaf_element.text = XML_ENCODERS.get(value.__class__, unicode)(value)
