# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3.
# See LICENSE comming with the source of 'mpuls offline'
# for details.
#
# author: Sascha L. Teichmann <sascha.teichmann@intevation.de>
#

import os
import traceback
import datetime
from cgi import parse_qs
import logging

import PyQt4.QtCore as QtCore
import PyQt4.QtGui as QtGui

from formed.meta.data import PageNode, RepeatNode
from formed.renderer.html import RepeatGroupRenderer, apply_hidden_booleans
from formed.instance.backends.xmlimport import XMLImportException

from lib.formedrenderer import FormData, UrlFactory, ViewRenderer, \
     ErrorRenderer
from lib.htmltemplate import HTML_TEMPLATE, EMPTY_HTML, HELP_TEMPLATE, FORM_MAIN
from lib.filecache import UTF8

from templates.logbook import LogbookOverviewRenderer, LogbookEntryRenderer, \
     NewLogbookEntryRenderer, EditLogbookEntryRenderer

from model.logbook import LogbookEntry, LogbookEntryErrors

from ui.dialogs import DefaultsDialog, AboutDialog, show_help
from ui.webview import WebView


log = logging.getLogger(__name__)


TITLE = "mpuls Offline-Klient"


class MainWindow(QtGui.QMainWindow):

    def __init__(self, main):
        QtGui.QMainWindow.__init__(self)

        self.setObjectName("python")

        self.resize(*self.initialSize())

        self.main = main
        self.store_directory = None

        self.setup_menu_items()
        self.web = WebView(self, main.file_cache)

        icon_path = main.file_cache.get_file_name(os.path.join('images',
                                                               'icons',
                                                               'app-icon.png'))

        self.setWindowIcon(QtGui.QIcon(icon_path))

        self.dump_html = (os.getenv('MPULS_OFFLINE_DUMP_HTML')
                          and self.__dump_html or self.__dump_null)

        self.wireWithHTML()
        self.setInitialPage()

        self.setCentralWidget(self.web)
        self.center()
        self.update_window_title()

    def setup_menu_items(self):
        new_case = QtGui.QAction('Neu', self)
        new_case.setShortcut('Ctrl+N')
        new_case.setStatusTip('Neue Akte anlegen')
        self.connect(new_case, QtCore.SIGNAL('triggered()'),
                     QtCore.SLOT('new_case()'))

        load_case = QtGui.QAction('Laden...', self)
        load_case.setShortcut('Ctrl+O')
        load_case.setStatusTip('Akte laden')
        self.connect(load_case, QtCore.SIGNAL('triggered()'),
                     QtCore.SLOT('load_case()'))

        save_case = QtGui.QAction('Speichern', self)
        save_case.setShortcut('Ctrl+S')
        save_case.setStatusTip('Akte speichern')
        self.connect(save_case, QtCore.SIGNAL('triggered()'),
                     QtCore.SLOT('save_case()'))

        save_as_case = QtGui.QAction('Speichern unter...', self)
        save_as_case.setShortcut('Ctrl+Shift+S')
        save_as_case.setStatusTip('Akte unter anderem Namen speichern')
        self.connect(save_as_case, QtCore.SIGNAL('triggered()'),
                     QtCore.SLOT('save_as_case()'))

        quit = QtGui.QAction('Beenden', self)
        quit.setShortcut('Ctrl+Q')
        quit.setStatusTip('Programm beenden')
        self.connect(quit, QtCore.SIGNAL('triggered()'),
                     QtCore.SLOT('close()'))

        if os.getenv('MPULS_OFFLINE_PRINT_PAGE'):
            printing = QtGui.QAction('Drucken', self)
            printing.setShortcut('Ctrl+P')
            printing.setStatusTip('Seite drucken...')
            self.connect(printing, QtCore.SIGNAL('triggered()'),
                         QtCore.SLOT('print_page()'))
        else:
            printing = None

        uniq_case = QtGui.QAction('Fallakte eindeutig machen', self)
        uniq_case.setStatusTip('Macht die interne Kennung der Akte eindeutig.')
        self.connect(uniq_case, QtCore.SIGNAL('triggered()'),
                     QtCore.SLOT('make_case_unique()'))

        menubar = self.menuBar()
        file = menubar.addMenu('&Datei')
        file.addAction(new_case)
        file.addAction(load_case)
        file.addAction(save_case)
        file.addAction(save_as_case)

        if printing:
            file.addSeparator()
            file.addAction(printing)

        file.addSeparator()

        file.addAction(uniq_case)

        file.addSeparator()

        file.addAction(quit)

        defaults = QtGui.QAction('Feldvorgaben...', self)
        defaults.setShortcut('Ctrl+E')
        defaults.setStatusTip(u'Vorgaben für Felder in der Akte')
        self.connect(defaults, QtCore.SIGNAL('triggered()'),
                     QtCore.SLOT('show_defaults_dialog()'))

        edit = menubar.addMenu('&Einstellungen')
        edit.addAction(defaults)

        about = QtGui.QAction(u'Über...', self)
        about.setStatusTip(u'Über den Offline-Klienten')
        self.connect(about, QtCore.SIGNAL('triggered()'),
                     QtCore.SLOT('about()'))

        help = menubar.addMenu('&Hilfe')
        help.addAction(about)

    def wireWithHTML(self):
        frame = self.web.page().mainFrame()
        self.connect(frame, QtCore.SIGNAL('javaScriptWindowObjectCleared()'),
                     QtCore.SLOT('register_javascript()'))

    @QtCore.pyqtSignature("void")
    def register_javascript(self):
        frame = self.web.page().mainFrame()
        frame.addToJavaScriptWindowObject("python", self)

    def isHTMLFormModified(self):
        """Return true if the values in the form of the HTML page have changed.
        """
        return self.call_javascript('isModified()').toBool()

    def initialSize(self):
        screen = QtGui.QDesktopWidget().screenGeometry()
        size = self.geometry()
        return (min(1024, screen.width()),
                min(718, max(10, screen.height()-50)))

    def center(self):
        screen = QtGui.QDesktopWidget().screenGeometry()
        size = self.geometry()
        self.move((screen.width() - size.width()) / 2,
                  (screen.height() - size.height()) / 2)

    def update_window_title(self):
        document = self.main.document

        if not document:
            title = TITLE
        else:
            filename = document.getFilename()
            filename = (not filename and "Kein Dateiname"
                        or os.path.basename(filename))
            dirty = document.isDirty() and " [modifiziert]" or ""
            title = u"%s - %s%s" % (TITLE, filename, dirty)

        self.setWindowTitle(title)

    def unsaved_data(self):
        return QtGui.QMessageBox.question(
            self, "Programm wirklich beenden?",
            u"Sie haben noch ungesicherte Daten.\n"
            u"Wollen Sie das Programm wirklich beenden?\n"
            u"Dabei könnten wichtige Änderungen verloren gehen.",
            QtGui.QMessageBox.Yes,
            QtGui.QMessageBox.No)

    def unsaved_new_data(self):
        return QtGui.QMessageBox.question(
            self,
            "Neue Akte wirklich anlegen?",
            u"Sie haben noch ungesicherte Daten.\n"
            u"Wollen Sie wirklich eine neue Akte anlegen?\n"
            u"Dabei könnten wichtige Änderungen verloren gehen.",
            QtGui.QMessageBox.Yes, QtGui.QMessageBox.No)

    def unsaved_load_data(self):
        return QtGui.QMessageBox.question(
            self,
            "Akte wirklich laden?",
            u"Sie haben noch ungesicherte Daten.\n"
            u"Wollen Sie wirklich eine andere Akte laden?\n"
            u"Dabei könnten wichtige Änderungen verloren gehen.",
            QtGui.QMessageBox.Yes, QtGui.QMessageBox.No)

    def ask_make_unique(self):
        return QtGui.QMessageBox.question(
            self,
            u"Akte wirklich eindeutig machen?",
            u"Wenn Sie eine Akten eindeutig machen, bekommt sie\n"
            u"intern eine neue Kennung. Mit dieser neuen Kennung\n"
            u"wird Sie beim Hochladen auf den zentralen Server\n"
            u"eindeutig und kollidiert nicht mehr mit anderen Akten.\n\n"
            u"Benutzen Sie diese Funktion nur, wenn Sie Fallakten\n"
            u"durch Kopieren von einander abgeleitet haben.\n\n"
            u"Wollen Sie die Akte wirklich eindeutig machen?",
            QtGui.QMessageBox.Yes, QtGui.QMessageBox.No)

    @QtCore.pyqtSignature("QCloseEvent")
    def closeEvent(self, evt):
        document = self.main.document
        if document and document.isDirty() or self.isHTMLFormModified():
            reply = self.unsaved_data()
            if reply == QtGui.QMessageBox.Yes:
                evt.accept()
            else:
                evt.ignore()
                return
        # cleanup swap-file
        swapfile = os.path.join(self.main.config.get_homedir(),
                                self.main.config.get_swapfile())
        try:
            os.remove(swapfile)
        except OSError, e:
            pass

        QtGui.QMainWindow.closeEvent(self, evt)

    @QtCore.pyqtSignature('void')
    def new_case(self):
        document = self.main.document
        if document and document.isDirty() or self.isHTMLFormModified():
            reply = self.unsaved_new_data()
            if reply == QtGui.QMessageBox.No:
                return
        self.main.new_document()
        self.setInitialPage()
        self.update_window_title()

    @QtCore.pyqtSignature('void')
    def load_case(self):
        document = self.main.document
        if document and document.isDirty() or self.isHTMLFormModified():
            reply = self.unsaved_load_data()
            if reply == QtGui.QMessageBox.No:
                return

        filename = QtGui.QFileDialog.getOpenFileName(
            self,
            "Bitte Dateinamen angeben.",
            self.get_store_directory(),
            "Fallakten (*.xml);;Sicherungsdateien (*.swp);;Letzte Versionen der Akten (*.bak)")

        if not filename:
            return

        swapfile = os.path.join(self.main.config.get_homedir(),
                                self.main.config.get_swapfile())
        filename = unicode(filename)

        self.set_store_directory(filename)

        try:
            self.main.load_document(filename)
        except XMLImportException, exc:
            QtGui.QMessageBox.critical(self, "Fehler beim Laden",
                                       (u"Die Fallakte konnte nicht geladen"
                                        u" werden wegen eines Problems mit der"
                                        u" XML-Datei:\n%s"
                                        % (exc.message,)))
            return
        except:
            QtGui.QMessageBox.critical(self, "Fehler beim Laden",
                                       (u"Beim Laden der Fallakte ist ein"
                                        u" Fehler aufgetreten:\n%s"
                                        % traceback.format_exc()))
            return

        self.setInitialPage()
        self.update_window_title()

    def get_store_directory(self):
        if not self.store_directory:
            self.store_directory = os.path.expanduser('~')
        return self.store_directory

    def set_store_directory(self, file):
        self.store_directory = os.path.dirname(file)

    @QtCore.pyqtSignature('void')
    def make_case_unique(self):
        document = self.main.document
        if document and self.ask_make_unique() == QtGui.QMessageBox.Yes:
            document.makeUnique()
            self.update_window_title()

    @QtCore.pyqtSignature('void')
    def save_case(self):
        document = self.main.document
        if not document or not document.isDirty():
            return

        filename = document.getFilename()

        if not filename:
            filename = QtGui.QFileDialog.getSaveFileName(
                self,
                "Bitte Dateinamen angeben.",
                self.get_store_directory(),
                "Fallakten (*.xml)")

            if not filename:
                return

            filename = unicode(filename)
            if not filename.lower().endswith(u'.xml'):
                filename += u'.xml'

            self.set_store_directory(filename)
            document.setFilename(filename)

        try:
            document.store()
        except:
            QtGui.QMessageBox.critical(self, "Fehler beim Abspeichern",
                                       (u"Beim Abspeichern der Fallakte ist"
                                        u" ein Fehler aufgetreten:\n%s"
                                        % traceback.format_exc()))
        self.update_window_title()

    @QtCore.pyqtSignature('void')
    def print_page(self):
        document = self.main.document
        if not document or not self.web:
            return
        printer = QtGui.QPrinter()
        dialog = QtGui.QPrintDialog(printer, self)
        if dialog.exec_() == QtGui.QDialog.Accepted:
            self.web.print_(printer)

    @QtCore.pyqtSignature('void')
    def save_as_case(self):
        document = self.main.document
        if not document:
            return

        filename = document.getFilename()

        if not filename:
            filename = self.get_store_directory()

        filename = QtGui.QFileDialog.getSaveFileName(
            self,
            "Bitte Dateinamen angeben.",
            filename,
            "Fallakten (*.xml)")

        if not filename:
            return

        filename = unicode(filename)
        if not filename.lower().endswith(u'.xml'):
            filename += u'.xml'

        self.set_store_directory(filename)
        document.setFilename(filename)

        try:
            document.store()
        except:
            QtGui.QMessageBox.critical(self, "Fehler beim Abspeichern",
                                       u"Beim Abspeichern der Fallakte ist"
                                       u" ein Fehler aufgetreten:\n%s"
                                       % traceback.format_exc())
        self.update_window_title()

    @QtCore.pyqtSignature('void')
    def about(self):
        about = AboutDialog(self, self.main.file_cache)
        about.setModal(True)
        about.exec_()

    @QtCore.pyqtSignature('void')
    def show_defaults_dialog(self):
        dialog = DefaultsDialog(self.main.formed, self.main.defaults, self)
        dialog.setModal(True)
        dialog.exec_()

    def render_empty_page(self):
        get_file = self.main.file_cache.get_file
        html = EMPTY_HTML.safe_substitute(
            ALL_CSS    = get_file("styles/all.css",    UTF8),
            COLOR_CSS  = get_file("styles/color3.css", UTF8),
            SCREEN_CSS = get_file("styles/screen.css", UTF8))

        self.dump_html(html)
        self.web.setHtml(html, QtCore.QUrl(self.main.base_url))
        self.setWindowTitle(TITLE)

    def parse_page_name(self, page_name):
        page_parts = page_name.rsplit('#', 1)
        if len(page_parts) > 1:
            page_name, anchor = page_parts[0:2]
        else:
            anchor = None

        if ":" in page_name:
            page_parts = page_name.split(":")
            meta_name = page_parts[0]
            rg_ids = [int(p) for p in page_parts[1:]]
        else:
            meta_name = page_name
            rg_ids = []

        return meta_name, rg_ids, page_name, anchor

    def convert_errors(self, items, page_name=None):
        """Extract messages and value from a dictionary with error items.
        The return value is a tuple with two dictionary, the first
        mapping formed item names to the message associated with the
        violated rul and the second mapping formed names to the
        associated value that violates the rule.

        If the page_name parameter is given and not false, only items on
        that page as determined by the document's page_id_of method are
        included in the return value.  Otherwise items from all pages
        are included.
        """
        errors = {}
        values = {}
        for items in items.itervalues():
            for e in items:
                if (not page_name
                    or page_name == self.main.document.page_id_of(e.name)):
                    errors[e.name] = e.msg
                    values[e.name] = e.bad
        return errors, values

    def get_page_errors(self, page_name):
        """Return the errors contained on the page given by page_name"""
        return self.convert_errors(self.main.document.getErrors(), page_name)

    def get_inconsistencies(self):
        """Return a dictionary mapping item names to inconsistency messages"""
        return self.convert_errors(self.main.document.get_inconsistencies())[0]

    def render_page(self, page_name, unfold=False):
        document = self.main.document

        if not page_name or not document:
            self.render_empty_page()
            return

        meta_name, rg_ids, page_name, anchor = self.parse_page_name(page_name)

        page = self.main.formed.findByClassAndName(PageNode, meta_name)
        if not page:
            log.warning("no page name named %r", meta_name)
            self.render_empty_page()
            return

        file_cache = self.main.file_cache

        # ideally, we should get the language from the environment
        # variables.  However, the whole application uses German in so
        # many places that we may as well hard-wire it for the
        # renderers, too.
        languages = ["de"]

        navigation = document.getNavigation()

        errors, error_values = self.get_page_errors(page_name)
        inconsistencies = self.get_inconsistencies()

        form_data = FormData(document.tree, error_values, errors,
                             inconsistencies)
        url_factory = UrlFactory(file_cache)

        if isinstance(page.getParent(), RepeatNode) and not rg_ids:
            form_renderer = RepeatGroupRenderer(form_data, page.getParent(),
                                                url_factory, read_only=False,
                                                languages=languages)
        else:
            if rg_ids:
                node = document.tree.get((rg_ids[0],
                                          page.getParent().getName()))
                form_data.push_node(node)
            form_renderer = ViewRenderer(form_data, page, url_factory,
                                         read_only=False, languages=languages)
        show_next = navigation.getNextPage(meta_name) is not None
        waska_form = form_renderer.render(show_next=show_next,
                                          hidden_fields=[("page", page_name)])

        warnings = form_renderer.warnings
        error_renderer = ErrorRenderer(form_data, self.main.formed,
                                       errors, warnings, {},
                                       inconsistencies, document,
                                       form_renderer.widget_order, url_factory,
                                       languages=languages)
        formular_errors = error_renderer.render()

        form_header = form_renderer.render_page_headings(all=True)

        ti = navigation.getTreeItem(meta_name)
        if ti and unfold:
            ti.unfold()

        navigation_html = navigation.render(meta_name, file_cache,
                                            document.getPageNamesWithErrors())

        main_content = FORM_MAIN.safe_substitute(FORM_HEADER=form_header,
                                                FORMULAR_ERRORS=formular_errors,
                                                 WASKA_FORM=waska_form)

        get_file = file_cache.get_file

        html = HTML_TEMPLATE.safe_substitute(NAVIGATION=navigation_html,
                                             MAIN_CONTENT=main_content)

        self.dump_html(html)

        url = self.main.base_url
        if anchor:
            url = self.main.base_url + meta_name + '#' + anchor
        self.web.setHtml(html, QtCore.QUrl(url))

    def __dump_null(self, html):
        pass

    def __dump_html(self, html):
        if not hasattr(self, 'dump_html_counter'):
            self.dump_html_counter = 0
        import codecs
        f = None
        try:
            f = codecs.open("html-dump-%03d.html" % self.dump_html_counter,
                            "w", "UTF-8")
            f.write(html)
        finally:
            if f:
                try:
                    f.close()
                except:
                    pass

        self.dump_html_counter += 1

    def first_page(self):
        for page in self.main.formed.findAllByClass(PageNode):
            return page.getName()
        return None

    def setInitialPage(self):
        self.render_page(self.first_page())

    def call_javascript(self, cmd):
        return self.web.page().mainFrame().evaluateJavaScript(cmd)

    def save_form_data(self, encoded_values):
        document = self.main.document
        if not document or not encoded_values:
            return None

        params = parse_qs(str(encoded_values), keep_blank_values=1)
        apply_hidden_booleans(params, value=("",))

        page = params.pop('page')[0]

        parsed_params = dict()
        for key, values in params.iteritems():
            assert len(values) == 1
            parsed_params[FormData.parse_form_name(key)] = values[0]

        document.save(parsed_params, page)

        return page

    def save_render(self, values, goto_next):
        page = self.save_form_data(values)
        if page is None:
            return

        if goto_next:
            page = self.get_next_page(page)

        self.render_page(page, goto_next)
        self.update_window_title()

    def get_next_page(self, page_id):
        """Return the form name of the next page after page_id"""
        document = self.main.document
        full_name = FormData.parse_form_name(page_id)
        page_name = full_name[-1]

        next_page_name = page_name
        next_page_ids = ()

        # in most cases, we can just use the next page from the
        # navigation items
        nav_next = document.getNavigation().getNextPage(page_name)
        if nav_next:
            next_page_name = nav_next

        # if the page is part of a repeat group, though, the next page
        # may be the next repeat group instance (pages in repeat groups
        # are currently direct descendants of repeat groups and a repeat
        # group instance will only contain that one page.
        if len(full_name) > 1:
            next_page_ids = full_name[:-2]

            page_node = self.main.formed.name2page[page_name]
            ids = document.tree.get(full_name[:-2]
                                    + (page_node.getParent().getName(),)).keys()
            for index, rg_id in enumerate(ids[:-1]):
                if rg_id == full_name[-2]:
                    next_page_ids += (ids[index + 1],)
                    next_page_name = page_name
                    break

        return FormData.form_name_from_full_name(next_page_ids
                                                 + (next_page_name,))

    @QtCore.pyqtSignature('QString')
    def save(self, page):
        self.save_render(page, False)
        self.call_javascript('changed = 0')

    @QtCore.pyqtSignature('QString')
    def save_next(self, page):
        self.save_render(page, True)

    @QtCore.pyqtSignature('QString')
    def select_item(self, page):
        self.render_page(unicode(page))

    @QtCore.pyqtSignature('QString')
    def create_repeat_group(self, rg_id):
        full_name = FormData.parse_form_name(unicode(rg_id))
        self.main.document.create_repeat_group(full_name)
        self.select_repeat_group_branch(full_name)

    @QtCore.pyqtSignature('QString')
    def delete_repeat_group(self, rg_id):
        full_name = FormData.parse_form_name(unicode(rg_id))
        self.main.document.delete_repeat_group(full_name)
        self.select_repeat_group_branch(full_name)

    def select_repeat_group_branch(self, full_name):
        # FIXME: does not handle nested repeat groups
        page = self.main.formed.findByName(full_name[-1]).getChildren()[0]
        self.select_branch(page.getName())

    @QtCore.pyqtSignature('QString, QString')
    def create_inline_repeat_group(self, params, rg_id):
        full_name = FormData.parse_form_name(unicode(rg_id))
        page = self.save_form_data(params)
        assert page is not None
        self.main.document.create_repeat_group(full_name)
        self.render_page(page, False)
        self.update_window_title()

    @QtCore.pyqtSignature('QString, QString')
    def delete_inline_repeat_group(self, params, rg_id):
        full_name = FormData.parse_form_name(unicode(rg_id))
        page = self.save_form_data(params)
        assert page is not None
        self.main.document.delete_repeat_group(full_name)
        self.render_page(page, False)
        self.update_window_title()

    @QtCore.pyqtSignature('QString, QString')
    def close_branch(self, page, old_page):
        document = self.main.document
        if not document:
            return
        navigation = document.getNavigation()
        ti = navigation.getTreeItem(page)
        if not ti:
            return
        ti.set_open(False)
        self.render_page(unicode(old_page))

    @QtCore.pyqtSignature('QString, QString')
    def open_branch(self, page, old_page):
        document = self.main.document
        if not document:
            return
        navigation = document.getNavigation()
        ti = navigation.getTreeItem(page)
        if not ti:
            return
        ti.set_open(True)
        self.render_page(unicode(old_page))

    @QtCore.pyqtSignature('QString')
    def select_branch(self, page):
        self.render_page(unicode(page), True)

    @QtCore.pyqtSignature('QString')
    def undo_error(self, item_id):
        full_name = FormData.parse_form_name(unicode(item_id))
        self.main.document.undo_pending_change(full_name)
        self.render_page(self.main.document.page_id_of(full_name))

    def show_help(self, itemname, kind, content_method):
        nc = self.main.formed.findByName(unicode(itemname))
        if nc is not None:
            show_help(HELP_TEMPLATE.safe_substitute(
                KIND    = kind,
                TITLE   = nc.getDescription(),
                CONTENT = getattr(nc, content_method)()), self)

    @QtCore.pyqtSignature('QString')
    def required_help(self, item):
        self.show_help(item, u"Erforderliches Feld", "getRequired")

    @QtCore.pyqtSignature('QString')
    def eval_help(self, item):
        self.show_help(item, u"Auswertungsrelevantes Feld", "getEvaluation")

    @QtCore.pyqtSignature('QString')
    def help(self, item):
        self.show_help(item, u"Hilfe", "getHelp")

    @QtCore.pyqtSignature('QString')
    def delete_logbook_entry(self, uuid):

        document = self.main.document
        if not document:
            return
        uuid = unicode(uuid)

        if document.deleteLogbookEntryByUUID(uuid):
            self.update_window_title()

        self.logbook_overview()

    def render_content(self, navigation_html, main_content):
        get_file = self.main.file_cache.get_file
        html = HTML_TEMPLATE.safe_substitute(NAVIGATION=navigation_html,
                                             MAIN_CONTENT=main_content)

        self.dump_html(html)
        self.web.setHtml(html, QtCore.QUrl(self.main.base_url))

    def get_default_navigation(self):
        document = self.main.document
        return document.getNavigation().render(self.first_page(),
                                               self.main.file_cache,
                                              document.getPageNamesWithErrors())


    @QtCore.pyqtSignature('void')
    def logbook_overview(self):
        document = self.main.document
        if not document:
            return

        main_content = LogbookOverviewRenderer().render(document.getLogbook(),
                                                        self.main.file_cache)

        self.render_content(self.get_default_navigation(), main_content)

    @QtCore.pyqtSignature('QString')
    def view_logbook_entry(self, uuid):
        document = self.main.document
        if not document:
            return

        logbook = document.getLogbook()
        uuid = unicode(uuid)
        entry = logbook.get_entry_by_uuid(uuid)

        if not entry:
            return

        self.render_content(self.get_default_navigation(),
                            LogbookEntryRenderer().render(entry))

    def render_new_logbook_entry(self, entry, errors, values):
        self.render_content(self.get_default_navigation(),
                            NewLogbookEntryRenderer().render(entry, errors,
                                                             values,
                                                        self.main.logbook_meta))

    @QtCore.pyqtSignature('void')
    def create_logbook_entry(self):
        if self.main.document:
            entry = LogbookEntry()
            entry.setDate(datetime.datetime.now())
            self.render_new_logbook_entry(entry, None, None)

    @QtCore.pyqtSignature('QString')
    def create_logbook_entry_store(self, values):
        values = str(values) # treat as bytes!
        values = parse_qs(values, keep_blank_values=1)
        errors = LogbookEntryErrors(values, self.main.logbook_meta)
        entry = errors.getEntry()
        if errors.hasErrors():
            self.render_new_logbook_entry(entry, errors, values)
        else:
            document = self.main.document
            if not document:
                return
            document.addEntryToLogbook(entry)
            self.logbook_overview()
            self.update_window_title()

    def render_edit_logbook_entry(self, entry, errors, values):
        self.render_content(self.get_default_navigation(),
                            EditLogbookEntryRenderer().render(entry, errors,
                                                              values,
                                                        self.main.logbook_meta))

    @QtCore.pyqtSignature('QString')
    def edit_logbook_entry(self, uuid):
        document = self.main.document
        if not document:
            return
        uuid = unicode(uuid)
        logbook = document.getLogbook()
        entry = logbook.get_entry_by_uuid(uuid)
        if entry:
            self.render_edit_logbook_entry(entry, None, None)

    @QtCore.pyqtSignature('QString')
    def edit_logbook_entry_store(self, values):
        values = str(values) # treat as bytes!
        values = parse_qs(values, keep_blank_values=1)
        errors = LogbookEntryErrors(values, self.main.logbook_meta)
        entry = errors.getEntry()
        if errors.hasErrors():
            self.render_edit_logbook_entry(entry, errors, values,
                                           self.main.logbook_meta)
        else:
            document = self.main.document
            if not document:
                return
            document.replaceLogbookEntryByUUID(entry)
            self.logbook_overview()
            self.update_window_title()

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
