# -*- coding: utf-8 -*-

"""Adapt libformed's HTML renderes for the offline client"""

import logging

from formed.renderer.html import HtmlRenderer, tag, \
     StructureFormData, UrlFactory as BaseUrlFactory, \
     ErrorRenderer as BaseErrorRenderer


log = logging.getLogger(__name__)


class FormData(StructureFormData):

    def __init__(self, instance, shadow, errors, inconsistencies):
        """Initialize the form data object.

        The parameter instance is the structure instance tree.  The
        parameter shadow is a dictionary mapping full instance item
        names to values which will be consulted before the instance tree
        when data is fetched, so that the values in the shadow dict
        shadow the ones from the instance tree.  The parameter errors is
        a dictionary mapping full item names to error messages
        associated with the item.  The parameter inconsistencies is a
        mapping analogous to the errors mapping, but contains messages
        for inconsistencies.
        """
        super(FormData, self).__init__(instance, shadow, errors)
        self.inconsistencies = inconsistencies

    def has_inconsistency(self, meta):
        return self.full_item_name(meta) in self.inconsistencies

    def get_inconsistencies(self, meta):
        message = self.inconsistencies.get(self.full_item_name(meta))
        if message:
            return (message,)
        return ()


class UrlFactory(BaseUrlFactory):

    url_map = {
        "/navigation/select_branch": "javascript:select_branch(\'%s\');",
        "/repeatgroup/new": "javascript:create_repeat_group(\'%s\');",
        "/repeatgroup/delete": "javascript:delete_repeat_group(\'%s\');",
        "/annotations/help": "javascript:help('%s');",
        "/annotations/required": "javascript:required_help('%s');",
        "/annotations/evaluation": "javascript:eval_help('%s');",
        }

    def __init__(self, file_cache):
        self.file_cache = file_cache

    def url_for(self, controller, action, id=None):
        url_key = controller + "/" + action

        if url_key == "/formularpage/undo":
            return self._undo_url(id)

        url = self.url_map.get(url_key)
        if url is not None:
            return url % (id,)

        log.warning("cannot create url for controller=%r, action=%r, id=%r",
                    controller, action, id)
        return ""

    def _undo_url(self, full_name):
        return ("javascript:undo_error('%s');"
                % (FormData.form_name_from_full_name(full_name),))

    def icon(self, name):
        try:
            return self.file_cache.get_image_url("images/icons/" + name)
        except IOError, e:
            log.warning("Cannot load icon %r", name)
            return name


class ViewRenderer(HtmlRenderer):

    def render_form_tag(self):
        return tag("form", id="waska_form", accept_charset="UTF-8")

    def render_submit_buttons(self, show_next):
        # FIXME: the code assumes that the page name is safe in JS and HTML
        page_name = self.view_root.getName()
        out = []
        out.append(tag("div", class_="widget formbuttons"))
        out.append(tag("input", type="button", name="save",
                       onclick="save_page('%s');" % page_name,
                       value=u"Übernehmen",
                       tabindex=self.next_tabindex()))
        if show_next:
            out.append("&nbsp;")
            out.append(tag("input", type="button", name="savenext",
                           onclick="save_next('%s');" % page_name,
                           value=u"Übernehmen und weiter",
                           tabindex=self.next_tabindex()))
        out.append('</div>')
        return out

    def render_inline_rg_delete_button(self, rg_id):
        out = [tag("a", href="javascript:delete_inline_rg('%s')" % rg_id),
               tag("img", src=self.icon("delete_active_22.png"),
                   border="0", width="16", height="16",
                   alt=u"Eintrag löschen", title=u"Eintrag löschen"),
               "</a>"]
        return "".join(out)

    def render_inline_rg_create_button(self, rg_id):
        out = [tag("a", href="javascript:create_inline_rg('%s')" % rg_id),
               tag("img", src=self.icon("new_16.png"),
                   border="0", width="16", height="16",
                   alt=u"Eintrag hinzufügen", title=u"Eintrag hinzufügen"),
               "</a>"]
        return "".join(out)

    def error_css_class(self, context, nc):
        if self.data.has_inconsistency(nc):
            return "error"
        return super(ViewRenderer, self).error_css_class(context, nc)

    def error_info(self, context, nc):
        messages = self.data.get_inconsistencies(nc)
        if messages:
            return ("error", messages)

        messages = self.data.get_errors(nc)
        if messages:
            return ("error", messages)

        messages = self.get_warnings(nc)
        if messages:
            return ("warning", messages)

        return ("", [])



class ErrorRenderer(BaseErrorRenderer):

    def __init__(self, data, meta_tree, errors, warnings, notices,
                 inconsistencies, document, widget_order, url_factory,
                 languages):
        super(ErrorRenderer, self).__init__(data, meta_tree, errors, warnings,
                                            notices, widget_order, url_factory,
                                            languages=languages)
        self.inconsistencies = inconsistencies
        self.document = document

    def render(self):
        def make_url(full_name):
            form_name = self.data.form_name_from_full_name(full_name)
            page_id = self.document.page_id_of(full_name)
            navi_item = self.document.navigation.getTreeItem(page_id)
            return navi_item.get_select_link("#f_%s" % form_name)

        return ("".join(self.render_items(self.inconsistencies,
                                          'Inkonsistenzen',
                                          self.icon("info/alert.png"),
                                          show_undo=False,
                                          item_url_function=make_url))
                + super(ErrorRenderer, self).render())
