# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3.
# See LICENSE comming with the source of 'mpuls offline'
# for details.
#
# author: Sascha L. Teichmann <sascha.teichmann@intevation.de>
#

import sys
import datetime
import time
from uuid import uuid4
from xml.sax import make_parser
import xml.etree.ElementTree as ET

from formed.meta.data import RepeatNode, TextLeaf, IntLeaf
from formed.meta.structure import StructureNode
from formed.instance.backends.xmlimport import ComponentFactory
from formed.instance.simplified import StructureInstanceNode, \
     StructureInstanceTree

from offline.lib.xmlhelper import StackedLoader, decode_int, \
     iso_datetime_string, decode_iso_datetime

from offline.lib.helpers import dd_mm_YYYY, HH_MM, parse_HH_MM, \
     parse_dd_mm_YYYY, parse_pos_int, safe_unicode


class EntryType(object):

    def __init__(self, description, value, is_editable):
        self.description = description
        self.value = value
        self.is_editable = is_editable

    def getDescription(self):
        return self.description

    def isEditable(self):
        return self.is_editable

    def getValue(self):
        return self.value


class LogbookMeta(object):

    def __init__(self, filename=None):
        self.entry_categories = []
        self.entry_types = []
        # current_category only used while parsing
        self.current_category = None
        if filename:
            self.load(filename)

    def default_start(self, name, attrs):
        pass

    def default_end(self, name, attrs, value):
        pass

    def entry_category_start(self, name, attrs):
        self.current_category = []
        self.entry_categories.append((attrs["name"], self.current_category))

    def entry_category_end(self, name, attrs, value):
        self.current_category = None

    def entry_type_end(self, name, attrs, text):
        value = attrs.get(u'value')
        if value is None:
            return
        editable = attrs.get('editable', 'true').lower() != 'false'
        entry_type = EntryType(text.strip(), int(value), editable)
        self.entry_types.append(entry_type)
        if self.current_category is not None:
            self.current_category.append(entry_type)

    def load(self, filename):
        factories = {
            u'entry-category': (self.entry_category_start,
                                self.entry_category_end),
            u'entry-type': (self.default_start, self.entry_type_end)
            }

        loader = StackedLoader()
        loader.add_factories(u'logbook-meta', factories,
                             (self.default_start, self.default_end))

        parser = make_parser()
        parser.setContentHandler(loader)

        f = open(filename.encode(sys.getfilesystemencoding()), "rb")
        try:
            parser.parse(f)
        finally:
            f.close()

    def find_entry_type_by_value(self, value):
        for entry_type in self.entry_types:
            if entry_type.value == value:
                return entry_type
        return None


class LogbookEntryErrors(object):

    def __init__(self, params, meta=None):
        entry = LogbookEntry()
        self.entry = entry
        self.errors = []

        def param(k):
            v = params.get(k, ('',))[0]
            # XXX: Python and the Holy Encoding Grail
            for enc in ('utf-8', 'latin-1'):
                try:
                    return unicode(v, enc).strip()
                except UnicodeDecodeError:
                    pass
            return safe_unicode(v).strip()

        time = param('time')

        h, m = 0, 0
        if time:
            try:
                h, m = parse_HH_MM(time)
            except ValueError, e:
                self.errors.append(('time', e.message))

        date = param('date')

        if not date:
            date = datetime.date.today().strftime("%d.%m.%Y")

        try:
            d, mon, y = parse_dd_mm_YYYY(date)
            try:
                entry.datum = datetime.datetime(y, mon, d, h, m)
            except ValueError:
                raise ValueError(u'Kein gültiges Datum')
        except ValueError, e:
            self.errors.append(('date', e.message))

        duration = param('duration')

        if duration:
            try:
                entry.setDuration(parse_pos_int(duration))
            except ValueError, e:
                self.errors.append(('duration', e.message))

        editor = param('editor')
        if editor:
            entry.setEditor(editor)

        short_notice = param('short_notice')
        if short_notice:
            entry.setShortNotice(short_notice)

        notice = param('notice')
        if notice:
            entry.setNotice(notice)

        uuid = param('uuid')
        if uuid:
            entry.setUUID(uuid)

        if meta:
            kind = decode_int(param('kind'))
            if not kind is None:
                kind = meta.find_entry_type_by_value(kind)
                if kind:
                    entry.setType(kind)

    def getEntry(self):
        return self.entry

    def hasErrors(self):
        return bool(self.errors)

    def hasError(self, field):
        for error in self.errors:
            if error[0] == field:
                return True
        return False

    def getErrors(self):
        return self.errors

def _cmp(a, b):
    if a is None and b is None:
        return 0
    if a is None:
        return -1
    if b is None:
        return +1
    return cmp(a, b)


def ORDER_ENTRIES_BY_DATE_ASC(a, b):
    return _cmp(a.datum, b.datum)


def ORDER_ENTRIES_BY_DATE_DESC(a, b):
    return _cmp(b.datum, a.datum)


class LogbookEntry(object):

    def __init__(self, uuid=None):
        self.uuid = uuid
        self.datum = None
        self.art = None
        self.bearbeiter = None
        self.kurz_notiz = None
        self.notiz = None
        self.dauer = None

    def getUUID(self):
        if not self.uuid:
            self.uuid = str(uuid4())
        return self.uuid

    def setUUID(self, uuid):
        self.uuid = uuid

    def setDate(self, datum):
        self.datum = datum

    def getDate(self):
        return dd_mm_YYYY(self.datum)

    def getTime(self):
        return HH_MM(self.datum)

    def setEditor(self, editor):
        self.bearbeiter = editor

    def getEditor(self):
        return u"" if self.bearbeiter is None else self.bearbeiter

    def getType(self):
        return self.art

    def setType(self, art):
        self.art = art

    def getDuration(self):
        return u"" if self.dauer is None else "%d" % self.dauer

    def setDuration(self, duration):
        self.dauer = duration

    def getShortNotice(self):
        return u"" if self.kurz_notiz is None else self.kurz_notiz

    def setShortNotice(self, short_notice):
        self.kurz_notiz = short_notice

    def getNotice(self):
        return u"" if self.notiz is None else self.notiz

    def setNotice(self, notice):
        self.notiz = notice


class Logbook(object):

    def __init__(self):
        self.entries = []

    def sorted_entries(self, cmp=ORDER_ENTRIES_BY_DATE_ASC):
        entries = self.entries[:]
        entries.sort(cmp)
        return entries

    def add_entry(self, entry):
        self.entries.append(entry)

    def delete_entry_by_uuid(self, uuid):
        for idx, entry in enumerate(self.entries):
            if entry.uuid == uuid:
                self.entries.pop(idx)
                return True
        return False

    def replace_entry_by_uuid(self, new_entry):
        nuuid = new_entry.uuid
        for idx, entry in enumerate(self.entries):
            if entry.uuid == nuuid:
                self.entries[idx] = new_entry
                return True
        return False

    def get_entry_by_uuid(self, uuid):
        for entry in self.entries:
            if entry.uuid == uuid:
                return entry
        return None

    def make_unique(self):
        """Assign new UUIDs to all entries"""
        for self.entry in self.entries:
            self.entry.setUUID(uuid4())


#
# Logbook import
#

def new_node(cls, name):
    node = cls()
    node.setName(name)
    return node

def get_logbook_import_factory():
    """Return the component factory for the logbook for the libformed xmlimport
    """
    entry_node = StructureNode(new_node(RepeatNode, "tagebuch-eintrag"),
                               [new_node(TextLeaf, "tagebuch-datum"),
                                new_node(TextLeaf, "tagebuch-bearbeiter"),
                                new_node(IntLeaf, "tagebuch-art"),
                                new_node(TextLeaf, "tagebuch-sonstiges"),
                                new_node(TextLeaf, "tagebuch-notiz"),
                                new_node(IntLeaf, "tagebuch-dauer"),

                                # Some applications have an additional field:
                                #new_node(IntLeaf, "tagebuch-typ"),
                                ],
                               [])
    return ComponentFactory(None,
                            StructureNode(None, [], [entry_node]),
                            StructureInstanceNode,
                            StructureInstanceTree)


def import_logbook(case_tree, logbook_tree, logbook_meta):
    """Import the logbook"""
    logbook = Logbook()

    entries = logbook_tree.root.get_repeat_groups("tagebuch-eintrag")
    for imported in entries.itervalues():
        entry = LogbookEntry()

        for key, value in imported.leaves.iteritems():
            if key == "tagebuch-art":
                entry_type = logbook_meta.find_entry_type_by_value(value)
                if entry_type:
                    entry.setType(entry_type)
            elif key == "tagebuch-dauer":
                entry.dauer = value
            elif key == "tagebuch-bearbeiter":
                entry.bearbeiter = value
            elif key == "tagebuch-notiz":
                entry.notiz = value
            elif key == "tagebuch-sonstiges":
                entry.kurz_notiz = value
            elif key == "tagebuch-datum":
                entry.datum = decode_iso_datetime(value.strip())
            else:
                raise ValueError("Unhandled item %r: %r" % (key, value))

        uuid = imported.uuid
        if not uuid:
            uuid = str(uuid4())
        entry.uuid = uuid

        logbook.entries.append(entry)

    return logbook


#
# Logbook export
#

def text_subelement(parent, tag, text):
    """Create an element under parent with text content.
    The text will always be converted to unicode, so it should either be
    unicode or be an object where the default conversion to unicode is
    appropriate for XML serialisation.
    """
    if text is not None:
        element = ET.SubElement(parent, tag)
        element.text = unicode(text)


def export_logbook(case, logbook):
    """Add the contents of the case with id case_id to the elementtree case"""
    logbook_element = ET.SubElement(case, "tagebuch")

    for entry in logbook.entries:
        entry_element = ET.SubElement(logbook_element, "tagebuch-eintrag")

        text_subelement(entry_element, "uuid_id", entry.uuid)
        text_subelement(entry_element, "tagebuch-datum",
                        iso_datetime_string(entry.datum))
        text_subelement(entry_element, "tagebuch-bearbeiter",
                        entry.bearbeiter)
        text_subelement(entry_element, "tagebuch-art",
                        entry.getType().getValue())
        text_subelement(entry_element, "tagebuch-sonstiges", entry.kurz_notiz)
        text_subelement(entry_element, "tagebuch-notiz", entry.notiz)
        if entry.dauer is not None:
            text_subelement(entry_element, "tagebuch-dauer", entry.dauer)


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
