# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3.
# See LICENSE comming with the source of 'mpuls offline'
# for details.
#
# author: Sascha L. Teichmann <sascha.teichmann@intevation.de>
#

import os.path
import shutil
from collections import defaultdict
from itertools import chain
from traceback import print_exc
import logging

from formed.meta.structure import StructureTree, structure_path
from formed.meta.semantic import SemanticError, checkAndConvert, ErrorItem
import formed.meta.data as data
from formed.instance.backends.xmlimport import XMLImporter, XMLImportException
from formed.instance.backends.xmlexport import serialize_case
from formed.instance.backends.common import evaluate, rule_available
from formed.instance.simplified import StructureVariableProvider, \
     StructureInstanceTree

from offline.lib.xmlhelper import decode_int, decode_date

from navigation import Navigation

from logbook import Logbook, get_logbook_import_factory, import_logbook, \
     export_logbook


log = logging.getLogger(__name__)


DECODERS = {
    data.RadioNode: decode_int,
    data.TextAreaLeaf: lambda s: s,
    data.TextLeaf: lambda s: s,
    data.ChoiceNode: decode_int,
    data.IntLeaf: decode_int,
    data.DateLeaf: decode_date
    }


class MemoryXMLImporter(XMLImporter):

    def import_from_file(self, filename):
        infile = open(filename, "rb")
        imported = None
        try:
            for case in self.read_cases_from_xml(infile):
                if imported is None:
                    imported = case
                else:
                    raise IOError("Only one case allowed in XML files")
        finally:
            infile.close()

        if imported is None:
            raise IOError("No case read from %r" % filename)

        return imported

    def version_supported(self, version):
        accepted = self.structure_tree.get_version()
        if accepted  != version:
            raise XMLImportException(u"Die XML-Datei hat die Version '%s',"
                                     u" unterstuetzt wird aber nur Version '%s'"
                                     % (version, accepted))


class MemoryDatastore(object):

    def __init__(self, formed, defaults, logbook_meta, filename=None):
        self.formed = formed
        self.structure_tree = StructureTree(formed)
        self.dirty = False
        self.defaults = defaults
        self.logbook_meta = logbook_meta
        self.navigation = None

        self.errors = {}

        if filename:
            self.load(filename)
        else:
            self.tree = StructureInstanceTree(self.structure_tree)
            self.tree.make_unique()
            self.logbook = Logbook()
            self.presetDefaults()

        self.check_inconsistencies()

        self.filename = filename
        self.swapfile = None

    def setSwapfile(self, swapfile=None):
        self.swapfile = swapfile

    def presetDefaults(self):
        defaults = self.defaults
        for item in defaults.values:
            try:
                widget = self.formed.name2nc[item.key]
            except KeyError:
                continue
            try:
                iv = item.eval()
            except:
                print_exc()
                continue
            decoder = DECODERS.get(widget.__class__, lambda s: s)
            try:
                iv = decoder(iv)
                self.tree.set_value(item.key, iv)
            except:
                print_exc()

    def isDirty(self):
        return self.dirty

    def setDirty(self, flag):
        self.dirty = flag

    def getFilename(self):
        return self.filename

    def getErrors(self):
        return self.errors

    def undo_pending_change(self, full_name):
        """Undo pending change of the item given by full_name"""
        if full_name in self.errors:
            del self.errors[full_name]

    def makeUnique(self):
        self.tree.make_unique()
        self.logbook.make_unique()
        self.dirty = True

    def page_id_of(self, full_name):
        """Return the page id (name + RG ids) containing the given item."""
        page_node = self.formed.name2page[full_name[-1]]
        num_page_rgs = len(structure_path(page_node)) - 1
        return ":".join([page_node.getName()]
                        + map(str, full_name[:num_page_rgs]))

    def getPageNamesWithErrors(self):
        return set([self.page_id_of(e.name)
                    for errors in self.errors.itervalues()
                    for e in errors])

    def getNavigation(self):
        if self.navigation is None:
            self.navigation = Navigation(self.formed)
        return self.navigation

    def setFilename(self, filename):
        if self.filename != filename:
            self.filename = filename
            self.dirty = True

    def getLogbook(self):
        if not self.logbook:
            self.logbook = Logbook()
        return self.logbook

    def deleteLogbookEntryByUUID(self, uuid):
        logbook = self.getLogbook()
        if logbook.delete_entry_by_uuid(uuid):
            self.dirty = True
            return True
        return False

    def replaceLogbookEntryByUUID(self, entry):
        logbook = self.getLogbook()
        if logbook.replace_entry_by_uuid(entry):
            self.dirty = True
            return True
        return False

    def addEntryToLogbook(self, entry):
        logbook = self.getLogbook()
        logbook.add_entry(entry)
        self.dirty = True

    def load(self, filename):
        importer = MemoryXMLImporter(self.formed,
                                     {"tagebuch": get_logbook_import_factory()})
        imported = importer.import_from_file(filename)
        self.tree = imported["master"]
        tagebuch = imported.get("tagebuch")
        if tagebuch is not None:
            self.logbook = import_logbook(self.tree, tagebuch,
                                          self.logbook_meta)
        else:
            self.logbook = Logbook()

        self.filename = filename
        self.dirty = False
        self.errors = dict()

    def get_inconsistencies(self):
        return self.inconsistencies

    def check_inconsistencies(self):
        self.inconsistencies = \
                             self.check_rules(dict(self.tree.iter_all_items()),
                                              find_best_changes=False)[-1]

    def backup_file(self, filename):
        if os.path.isfile(filename):
            try:
                shutil.copy(filename, filename + ".bak")
            except OSError, e:
                log.exception("Could not create backup file for %r",
                              filename)

    def store(self, filename=None, backup=True):
        if filename is None:
            if not self.filename:
                return
            filename = self.filename

        # Write a backup file before storing data. The backup should always
        # contain the data of the last save
        if backup:
            self.backup_file(filename)

        def serialize_logbook(case):
            export_logbook(case, self.logbook)

        outfile = open(filename, "w")
        try:
            outfile.write(serialize_case(self.tree, external=serialize_logbook))
            if backup:
                self.dirty = False
        finally:
            outfile.close()

    def save(self, params, page_name):
        new_values = dict()

        old_errors = self.errors.copy()

        new_errors = defaultdict(list)

        for key, value in params.iteritems():
            old_errors.pop(key, None)
            widget = self.formed.name2nc[key[-1]]
            try:
                new_values[key] = (checkAndConvert(widget, value) if value
                                   else None)
            except SemanticError, err:
                new_errors[key].append(ErrorItem(ErrorItem.SEMANTIC,
                                                 key, err.value, value))

        changes, eval_errors = self.check_rules(new_values)

        # merge errors to form the new errors dict
        for key, errors in chain(eval_errors.iteritems(),
                                 old_errors.iteritems()):
            new_errors[key].extend(errors)

        self.errors = new_errors

        for key, value in changes:
            self.tree.set_value(key, value)
            self.dirty = True

        self.check_inconsistencies()

        # Write swap file, but do not overwrite the backup
        try:
            self.store(self.swapfile, backup=False)
        except:
            log.exception("Error while trying to write swapfile %r",
                          self.swapfile)

    def check_rules(self, new_values, find_best_changes=True):
        return evaluate(new_values, self.formed, self.rule_available,
                        self.create_eval_context, lambda k: k[-1],
                        find_best_changes=find_best_changes)

    def rule_available(self, rule, eval_context):
        return rule_available(self.formed, rule, eval_context)

    def create_eval_context(self, key, shadow):
        return StructureVariableProvider(self.tree, key, shadow)

    def create_repeat_group(self, repeat_group):
        if len(repeat_group) == 1:
            # we can only handle non-nested repeat groups at the moment.
            self.tree.root.create_repeat_group(repeat_group[-1])

    def delete_repeat_group(self, repeat_group):
        if len(repeat_group) == 2:
            # we can only handle non-nested repeat groups at the moment.
            rg_id, rg_name = repeat_group
            self.tree.root.delete_repeat_group(rg_name, rg_id)

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
